package org.asciidoc.intellij.quickfix;

import com.intellij.codeInspection.LocalQuickFix;
import com.intellij.codeInspection.ProblemDescriptor;
import com.intellij.codeInspection.util.IntentionFamilyName;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.TextRange;
import com.intellij.psi.PsiElement;
import com.intellij.psi.ResolveResult;
import com.intellij.psi.impl.source.tree.LeafPsiElement;
import org.asciidoc.intellij.AsciiDocBundle;
import org.asciidoc.intellij.lexer.AsciiDocTokenTypes;
import org.asciidoc.intellij.psi.AsciiDocBlockId;
import org.asciidoc.intellij.psi.AsciiDocFileReference;
import org.asciidoc.intellij.psi.AsciiDocLink;
import org.asciidoc.intellij.psi.AsciiDocSectionImpl;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Objects;

/**
 * @author Alexander Schwartz 2020
 */
public class AsciiDocChangeXrefWithNaturalCrossReferenceToId implements LocalQuickFix {

  @Override
  public @IntentionFamilyName @NotNull String getFamilyName() {
    return AsciiDocBundle.message("asciidoc.quickfix.changeXrefWithNaturalCrossReferenceToId");
  }

  @Override
  public void applyFix(@NotNull Project project, @NotNull ProblemDescriptor descriptor) {
    PsiElement element = descriptor.getPsiElement();
    if (element instanceof AsciiDocLink) {
      AsciiDocLink link = (AsciiDocLink) element;
      LeafPsiElement child = getChild(link);
      if (child != null) {
        AsciiDocFileReference fileReference = link.getAnchorReference();
        if (fileReference != null) {
          AsciiDocSectionImpl section = getSection(link);
          if (section != null) {
            String id;
            AsciiDocBlockId blockId = section.getBlockId();
            if (blockId != null) {
              id = blockId.getName();
              Objects.requireNonNull(id);
            } else {
              id = section.getAutogeneratedId();
            }
            child.replaceWithText(fileReference.getRangeInElement().shiftLeft(child.getStartOffsetInParent()).replace(child.getText(), id));
          }
        }
      }
    }
  }

  public boolean canFix(AsciiDocLink link) {
    return getChild(link) != null && getSection(link) != null;
  }

  @Nullable
  private LeafPsiElement getChild(AsciiDocLink link) {
    AsciiDocFileReference fileReference = link.getAnchorReference();
    if (fileReference != null) {
      ResolveResult[] resolveResultsFile = fileReference.multiResolve(false);
      if (resolveResultsFile.length == 1) {
        PsiElement child = link.getFirstChild();
        TextRange range = fileReference.getRangeInElement();
        while (child != null && range.getStartOffset() >= child.getTextLength()) {
          range = range.shiftRight(-child.getTextLength());
          child = child.getNextSibling();
        }
        if (child instanceof LeafPsiElement && range.getEndOffset() <= child.getTextLength()
          && child.getNode().getElementType() == AsciiDocTokenTypes.LINKFILE) {
          return (LeafPsiElement) child;
        }
      }
    }
    return null;
  }

  @Nullable
  private AsciiDocSectionImpl getSection(AsciiDocLink link) {
    AsciiDocFileReference fileReference = link.getAnchorReference();
    if (fileReference != null) {
      ResolveResult[] resolveResultsFile = fileReference.multiResolve(false);
      if (resolveResultsFile.length == 1 && resolveResultsFile[0].getElement() instanceof AsciiDocSectionImpl) {
        return ((AsciiDocSectionImpl) resolveResultsFile[0].getElement());
      }
    }
    return null;
  }

}
